/* $Header: /home/cvs/bp/oofem/oofemlib/src/fei2dquadlin.C,v 1.2.4.1 2004/04/05 15:19:43 bp Exp $ */
/*

                   *****    *****   ******  ******  ***   ***                            
                 **   **  **   **  **      **      ** *** **                             
                **   **  **   **  ****    ****    **  *  **                              
               **   **  **   **  **      **      **     **                               
              **   **  **   **  **      **      **     **                                
              *****    *****   **      ******  **     **         
            
                                                                   
               OOFEM : Object Oriented Finite Element Code                 
                    
                 Copyright (C) 1993 - 2000   Borek Patzak                                       



         Czech Technical University, Faculty of Civil Engineering,
     Department of Structural Mechanics, 166 29 Prague, Czech Republic
                                                                               
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                                                                              
*/

#include "fei2dquadlin.h"
#include "flotarry.h"
#include "flotmtrx.h"
#include "intarray.h"
#include "node.h"
#include "mathfem.h"

void
FEI2dQuadLin::evalN (FloatArray& answer, const FloatArray& lcoords, double time)
{
 double ksi, eta;
 answer.resize (4);
 
 ksi = lcoords.at(1);
 eta = lcoords.at(2);

 answer.at(1) = (1. + ksi) * (1. + eta) * 0.25 ;
 answer.at(2) = (1. - ksi) * (1. + eta) * 0.25 ;
 answer.at(3) = (1. - ksi) * (1. - eta) * 0.25 ;
 answer.at(4) = (1. + ksi) * (1. - eta) * 0.25 ;

 return;
}

void
FEI2dQuadLin :: evaldNdx (FloatMatrix&answer, Domain* d, IntArray& nodes, const FloatArray& lcoords, double time)
{
 int i;
 FloatMatrix jacobianMatrix(2,2), inv(2,2);
 FloatArray nx(4), ny(4);

 this -> giveJacobianMatrixAt (jacobianMatrix, d, nodes, lcoords);
 inv.beInverseOf (jacobianMatrix);

 this->giveDerivativeKsi(nx, lcoords.at(2));
 this->giveDerivativeEta(ny, lcoords.at(1));

 answer.resize(4,2);

 for (i=1;i<=4;i++) {
  answer.at(i,1) = nx.at(i)*inv.at(1,1) + ny.at(i)*inv.at(1,2);
  answer.at(i,2) = nx.at(i)*inv.at(2,1) + ny.at(i)*inv.at(2,2);
 }
}

void
FEI2dQuadLin :: local2global (FloatArray& answer, Domain* d, IntArray& nodes, const FloatArray& lcoords, double time)
{
 double ksi, eta, n1, n2, n3, n4;
 
 ksi = lcoords.at(1) ;
 eta = lcoords.at(2) ;
 
 n1 = (1. + ksi) * (1. + eta) * 0.25 ;
 n2 = (1. - ksi) * (1. + eta) * 0.25 ;
 n3 = (1. - ksi) * (1. - eta) * 0.25 ;
 n4 = (1. + ksi) * (1. - eta) * 0.25 ;
 
 answer.resize(2);
 answer.at(1) = n1*d->giveNode(nodes.at(1))->giveCoordinate(xind) + n2*d->giveNode(nodes.at(2))->giveCoordinate(xind)+
  n3 * d->giveNode(nodes.at(3))->giveCoordinate(xind) + n4 * d->giveNode(nodes.at(4))->giveCoordinate(xind);
 answer.at(2) = n1*d->giveNode(nodes.at(1))->giveCoordinate(yind) + n2*d->giveNode(nodes.at(2))->giveCoordinate(yind)+
  n3 * d->giveNode(nodes.at(3))->giveCoordinate(yind) + n4 * d->giveNode(nodes.at(4))->giveCoordinate(yind);
}

#define POINT_TOL 1.e-6

int
FEI2dQuadLin :: global2local (FloatArray& answer, Domain* d, IntArray& nodes, const FloatArray& coords, double time)
{
   Node    *node1,*node2,*node3,*node4;
   double  x1,x2,x3,x4,y1,y2,y3,y4,a1,a2,a3,a4,b1,b2,b3,b4;
  double a,b,c, ksi1, ksi2, ksi3, eta1=0.0, eta2=0.0, denom;
  int nroot;

  answer.resize(2);

   node1 = d -> giveNode(nodes.at(1)) ;
   node2 = d -> giveNode(nodes.at(2)) ;
   node3 = d -> giveNode(nodes.at(3)) ;
   node4 = d -> giveNode(nodes.at(4)) ;

   x1 = node1 -> giveCoordinate(xind) ;
   x2 = node2 -> giveCoordinate(xind) ;
   x3 = node3 -> giveCoordinate(xind) ;
   x4 = node4 -> giveCoordinate(xind) ;

   y1 = node1 -> giveCoordinate(yind) ;
   y2 = node2 -> giveCoordinate(yind) ;
   y3 = node3 -> giveCoordinate(yind) ;
   y4 = node4 -> giveCoordinate(yind) ;

  a1 = x1+x2+x3+x4;
  a2 = x1-x2-x3+x4;
  a3 = x1+x2-x3-x4;
  a4 = x1-x2+x3-x4;

  b1 = y1+y2+y3+y4;
  b2 = y1-y2-y3+y4;
  b3 = y1+y2-y3-y4;
  b4 = y1-y2+y3-y4;

  a = a2*b4-b2*a4;
  b = a1*b4+a2*b3-a3*b2-b1*a4-b4*4.0*coords.at(xind)+a4*4.0*coords.at(yind);
  c = a1*b3-a3*b1-4.0*coords.at(xind)*b3+4.0*coords.at(yind)*a3;

  // solve quadratic equation for ksi
  cubic (0.0, a, b, c, &ksi1, &ksi2, &ksi3, &nroot);

  if (nroot == 0) return 0;
  if(nroot){
   denom = (b3+ksi1*b4);
   if (fabs(denom) <= 1.0e-10) 
    eta1 = (4.0*coords.at(xind)-a1-ksi1*a2) / (a3+ksi1*a4);
   else eta1 = (4.0*coords.at(yind)-b1-ksi1*b2) / denom;
  }

  if(nroot > 1){
   double diff_ksi1, diff_eta1, diff_ksi2, diff_eta2, diff1, diff2;

   denom = b3+ksi2*b4;
   if (fabs(denom) <= 1.0e-10) 
    eta2 = (4.0*coords.at(xind)-a1-ksi2*a2) / (a3+ksi2*a4);
   else eta2 = (4.0*coords.at(yind)-b1-ksi2*b2) / denom;
   
   // choose the one which seems to be closer to the parametric space (square <-1;1>x<-1;1>)
   diff_ksi1 = 0.0;
   if(ksi1 > 1.0)diff_ksi1 = ksi1 - 1.0;
   if(ksi1 < -1.0)diff_ksi1 = ksi1 + 1.0;

   diff_eta1 = 0.0;
   if(eta1 > 1.0)diff_eta1 = eta1 - 1.0;
   if(eta1 < -1.0)diff_eta1 = eta1 + 1.0;
   
   diff_ksi2 = 0.0;
   if(ksi2 > 1.0)diff_ksi2 = ksi2 - 1.0;
   if(ksi2 < -1.0)diff_ksi2 = ksi2 + 1.0;

   diff_eta2 = 0.0;
   if(eta2 > 1.0)diff_eta2 = eta2 - 1.0;
   if(eta2 < -1.0)diff_eta2 = eta2 + 1.0;
   
   diff1 = diff_ksi1 * diff_ksi1 + diff_eta1 * diff_eta1;
   diff2 = diff_ksi2 * diff_ksi2 + diff_eta2 * diff_eta2;

   // ksi2, eta2 seems to be closer
   if(diff1 > diff2){
    ksi1 = ksi2;
    eta1 = eta2;
   }
  }

  answer.at(1) = ksi1;
  answer.at(2) = eta1;

  // test if inside
  for (int i=1; i<=2; i++) {
   if (answer.at(i)<(-1.-POINT_TOL)) return 0;
   if (answer.at(i)>(1.+POINT_TOL)) return 0;
  }

  return 1;
}

double
FEI2dQuadLin :: giveTransformationJacobian (Domain* d, IntArray& nodes, const FloatArray& lcoords, double time)
{
 FloatMatrix jacobianMatrix (2,2);

 this -> giveJacobianMatrixAt (jacobianMatrix, d, nodes, lcoords);
 return jacobianMatrix.giveDeterminant();
}

void 
FEI2dQuadLin :: edgeEvalN (FloatArray& answer, const FloatArray& lcoords, double time)
{
 double ksi = lcoords.at(1);
 answer.resize(2);
 
 answer.at(1) = (1. - ksi) * 0.5 ;
 answer.at(2) = (1. + ksi) * 0.5 ;
}

void 
FEI2dQuadLin :: edgeEvaldNdx (FloatMatrix&answer, int iedge, 
              Domain* d, IntArray& nodes, const FloatArray& lcoords, double time)
{
 double l;
 IntArray edgeNodes;
 this->computeEdgeMapping (edgeNodes, nodes, iedge);
 l = this->edgeComputeLength(edgeNodes, d);

 answer.resize (2,1);
 answer.at(1,1) = -1.0/l;
 answer.at(2,1) =  1.0/l;
}

void
FEI2dQuadLin :: edgeLocal2global (FloatArray& answer, int iedge, 
                 Domain* d, IntArray& nodes, const FloatArray& lcoords, double time) 
{
 IntArray edgeNodes;
 FloatArray n;
 this->computeEdgeMapping (edgeNodes, nodes, iedge);
 this->edgeEvalN (n, lcoords, time);
 
 answer.resize(2);
 answer.at(1) = (n.at(1)* d->giveNode(edgeNodes.at(1))->giveCoordinate(xind) +
         n.at(2)* d->giveNode(edgeNodes.at(2))->giveCoordinate(xind));
 answer.at(2) = (n.at(1)* d->giveNode(edgeNodes.at(1))->giveCoordinate(yind) +
         n.at(2)* d->giveNode(edgeNodes.at(2))->giveCoordinate(yind));
 
}


double
FEI2dQuadLin :: edgeGiveTransformationJacobian (int iedge, Domain* d, IntArray& nodes, const FloatArray& lcoords, double time) 
{
 IntArray edgeNodes;
 this->computeEdgeMapping (edgeNodes, nodes, iedge);
 return 0.5 * this->edgeComputeLength(edgeNodes, d);
}


void 
FEI2dQuadLin :: computeEdgeMapping (IntArray& edgeNodes, IntArray& elemNodes, int iedge)
{
 int aNode=0, bNode=0;
 edgeNodes.resize(2);
 
  if (iedge == 1) { // edge between nodes 1 2
  aNode = elemNodes.at(1);
  bNode = elemNodes.at(2);
  } else if (iedge == 2) { // edge between nodes 2 3
  aNode = elemNodes.at(2);
  bNode = elemNodes.at(3);
  } else if (iedge == 3) { // edge between nodes 3 4
  aNode = elemNodes.at(3);
  bNode = elemNodes.at(4);
  } else if (iedge == 4) { // edge between nodes 4 1
  aNode = elemNodes.at(4);
  bNode = elemNodes.at(1);
  } else {
  OOFEM_ERROR2 ("FEI2dQuadLin :: computeEdgeMapping: wrong egde number (%d)", iedge);
  }
 
 edgeNodes.at(1) = aNode;
 edgeNodes.at(2) = bNode;
}

double
FEI2dQuadLin :: edgeComputeLength (IntArray& edgeNodes, Domain* d)
{
 double dx,dy;
  Node   *nodeA,*nodeB ;

  nodeA   = d->giveNode(edgeNodes.at(1)) ;
  nodeB   = d->giveNode(edgeNodes.at(2)) ;
 
  dx      = nodeB->giveCoordinate(xind) - nodeA->giveCoordinate(xind) ;
  dy      = nodeB->giveCoordinate(yind) - nodeA->giveCoordinate(yind) ;
  return (sqrt(dx*dx + dy*dy));
}

void
FEI2dQuadLin :: giveJacobianMatrixAt (FloatMatrix& jacobianMatrix, Domain* d, const IntArray& nodes, const FloatArray& lcoords)
// Returns the jacobian matrix  J (x,y)/(ksi,eta)  of the receiver.
// Computes it if it does not exist yet.
{
 int        i;
 double     ksi,eta,x,y;
 FloatArray nx(4),ny(4);

 jacobianMatrix.resize(2,2);
 jacobianMatrix.zero();
 
 ksi = lcoords.at(1);
 eta = lcoords.at(2);

 this->giveDerivativeKsi(nx, eta);
 this->giveDerivativeEta(ny, ksi);

 for (i=1;i<=4;i++){
  x = d -> giveNode(nodes.at(i)) -> giveCoordinate(xind);
  y = d -> giveNode(nodes.at(i)) -> giveCoordinate(yind);

  jacobianMatrix.at(1,1) += nx.at(i)*x;
  jacobianMatrix.at(1,2) += nx.at(i)*y;
  jacobianMatrix.at(2,1) += ny.at(i)*x;
  jacobianMatrix.at(2,2) += ny.at(i)*y;
 }
}


void
FEI2dQuadLin :: giveDerivativeKsi (FloatArray& n, double eta)
{
 n.resize(4);

 n.at(1) =  0.25*(1.+eta);
 n.at(2) = -0.25*(1.+eta);
 n.at(3) = -0.25*(1.-eta);
 n.at(4) =  0.25*(1.-eta);
}

void
FEI2dQuadLin :: giveDerivativeEta (FloatArray& n, double ksi)
{
 n.resize(4);

 n.at(1) =  0.25*(1.+ksi);
 n.at(2) =  0.25*(1.-ksi);
 n.at(3) = -0.25*(1.-ksi);
 n.at(4) = -0.25*(1.+ksi);

}

